/*
 *  inode.c
 *
 *  Copyright (C) 1995,1996 Martin von Lwis
 *  Copyright (C) 1996 Albert D. Cahalan
 *  Copyright (C) 1996 Regis Duchesne
 */

#include <errno.h>
#include "ntfs.h"
#include "config.h"

int ntfs_init_inode(ntfs_inode *ino,ntfs_volume* vol,int inum)
{
	ntfs_debug("Initializing inode %x\n",inum);
	ino->i_number=inum;
	ino->vol=vol;
	ino->attr=ntfs_malloc(vol->mft_recordsize);
	if(ntfs_read_mft_record(vol,inum,ino->attr)==-1){
		ntfs_debug("init inode: %x failed\n",inum);
		return -1;
	}
	ntfs_debug("Init: got mft %x\n",inum);
	ino->attr_count=0;
	ino->record_count=0;
	ino->records=0;
	ino->attrs=0;
	ntfs_load_attributes(ino);
	ntfs_debug("Init: done %x\n",inum);
	return 0;
}

void ntfs_clear_inode(ntfs_inode *ino)
{
	int i;
	ntfs_free(ino->attr);
	ntfs_free(ino->records);
	for(i=0;i<ino->attr_count;i++)
	{
		if(ino->attrs[i].name)
			ntfs_free(ino->attrs[i].name);
		if(RESIDENT(&ino->attrs[i].header))
		{
			if(ino->attrs[i].d.data)
				ntfs_free(ino->attrs[i].d.data);
		}else{
			if(ino->attrs[i].d.r.runlist)
				ntfs_free(ino->attrs[i].d.r.runlist);
		}
	}
	ntfs_free(ino->attrs);
}

/* Insert all attributes from the record mftno of the MFT in the inode ino */
void ntfs_insert_mft_attributes(ntfs_inode* ino,char *mft,int mftno)
{
	int i;
	char *it;
	int type,len;
	/* check for duplicate */
	for(i=0;i<ino->record_count;i++)
		if(ino->records[i]==mftno)
			return;
	/* (re-)allocate space if necessary */
	if(ino->record_count % 8==0)
	{
		int *old=ino->records;
		ino->records=ntfs_malloc((ino->record_count+8)*sizeof(int));
		if(old) {
			for(i=0;i<ino->record_count;i++)
				ino->records[i]=old[i];
			ntfs_free(old);
		}
	}
	ino->records[ino->record_count]=mftno;
	ino->record_count++;
	it = mft + NTFS_GETU16(mft + 0x14);
	do{
		type=NTFS_GETU32(it);
		len=NTFS_GETU32(it+4);
		if(type!=-1)
			ntfs_insert_attribute(ino,it);
		it+=len;
	}while(type!=-1); /* attribute list ends with type -1 */
}

/* Read and insert all the attributes of an 'attribute list' attribute */
static int parse_attributes(ntfs_inode* ino,void *alist,int len)
{
	char *mft;
	int mftno,l;
	int last_mft=-1;
	mft=ntfs_malloc(ino->vol->mft_recordsize);
	while(len>0)
	{
	        /* process an attribute description */
		mftno=NTFS_GETU32(alist+0x10); /* BUG: this is u64 */
		if(mftno!=last_mft){
			last_mft=mftno;
			/* FIXME: avoid loading record if it's 
			   already processed */
			if(ntfs_read_mft_record(ino->vol,mftno,mft)==-1)
				return -1;
			ntfs_insert_mft_attributes(ino,mft,mftno);
		}
		l=NTFS_GETU16(alist+4);
		len-=l;
		alist+=l;
	}
	ntfs_free(mft);
	return 0;
}

void ntfs_load_attributes(ntfs_inode* ino)
{
	ntfs_attribute *alist;
	int datasize;
	int offset,len;
	char *buf;
	ntfs_debug("load_attributes %x 1\n",ino->i_number);
	ntfs_insert_mft_attributes(ino,ino->attr,ino->i_number);
	ntfs_debug("load_attributes %x 2\n",ino->i_number);
	alist=ntfs_find_attr(ino,AT_ATTRIBUTE_LIST,0);
	ntfs_debug("load_attributes %x 3\n",ino->i_number);
	if(!alist)
		return;
	ntfs_debug("load_attributes %x 4\n",ino->i_number);
	datasize=DATASIZE(&alist->header);
	if(RESIDENT(&alist->header))
	{
		parse_attributes(ino,alist->d.data,datasize);
		return;
	}
	buf=ntfs_malloc(1024);
	for(offset=0;datasize;datasize-=len)
	{
		ntfs_io io;
		io.fn_put=ntfs_put;
		io.fn_get=0;
		io.param=buf;
		len=min(datasize,1024);
		/* FIXME: this is really broken if there the attribute list
		   is larger than 1024 bytes and 1024 bytes is not a
		   attribute record border */
		ntfs_read_attr(ino,AT_ATTRIBUTE_LIST,0,offset,&io,len);
		parse_attributes(ino,buf,len);
	}
	ntfs_debug("load_attributes %x 5\n",ino->i_number);
	ntfs_free(buf);
}
	
/* Check and fixup a MFT record */
int ntfs_check_mft_record(ntfs_volume *vol,char *record)
{
	return ntfs_fixup_record(vol, record, "FILE", vol->mft_recordsize);
}

/* Return (in result) the value indicating the next available attribute 
   chunk number. Works for inodes w/o extension records only */
int ntfs_allocate_attr_number(ntfs_inode *ino, int *result)
{
	if(ino->record_count!=1)
		return EOPNOTSUPP;
	*result=NTFS_GETU16(ino->attr+0x28);
	NTFS_PUTU16(ino->attr+0x28, (*result)+1);
	return 0;
}

/* find the location of an attribute in the inode. A name of NULL indicates
   unnamed attributes. Return pointer to attribute or NULL if not found */
char *ntfs_get_attr(ntfs_inode *ino,int attr,char *name)
{
	/* location of first attribute */
	char *it= ino->attr + NTFS_GETU16(ino->attr + 0x14);
	int type;
	int len;
	/* Only check for magic DWORD here, fixup should have happened before */
	if(!IS_MFT_RECORD(ino->attr))return 0;
	do{
		type=NTFS_GETU32(it);
		len=NTFS_GETU16(it+4);
		/* We found the attribute type. Is the name correct, too? */
		if(type==attr)
		{
			int namelen=NTFS_GETU8(it+9);
			char *name_it;
			/* match given name and attribute name if present,
			   make sure attribute name is Unicode */
			for(name_it=it+NTFS_GETU16(it+10);namelen;
			    name++,name_it+=2,namelen--)
				if(*name_it!=*name || name_it[1])break;
			if(!namelen)break;
		}
		it+=len;
	}while(type!=-1); /* attribute list end with type -1 */
	if(type==-1)return 0;
	return it;
}

/* Search in an inode an attribute by type and name */
ntfs_attribute* ntfs_find_attr(ntfs_inode *ino,int type,char *name)
{
	int i;
	if(!ino){
		ntfs_error("ntfs_find_attr: NO INODE!\n");
		return 0;
	}
	for(i=0;i<ino->attr_count;i++)
	{
		if(type==ino->attrs[i].type)
		{
			if(!name && !ino->attrs[i].name)
				return ino->attrs+i;
			if(name && !ino->attrs[i].name)
				return 0;
			if(!name && ino->attrs[i].name)
				return 0;
			if(ntfs_ua_strncmp(ino->attrs[i].name,name,ntfs_strlen(name))==0)
				return ino->attrs+i;
		}
		if(type<ino->attrs[i].type)
			return 0;
	}
	return 0;
}

int ntfs_get_attr_size(ntfs_inode*ino,int type,char*name)
{
	ntfs_attribute *attr=ntfs_find_attr(ino,type,name);
	if(!attr)return 0;
	return DATASIZE(&attr->header);
}
	
int ntfs_attr_is_resident(ntfs_inode*ino,int type,char*name)
{
	ntfs_attribute *attr=ntfs_find_attr(ino,type,name);
	if(!attr)return 0;
	return RESIDENT(&attr->header);
}
	
/*
 * A run is coded as a type indicator, an unsigned length, and a signed cluster
 * offset.
 * . To save space, length and offset are fields of variable length. The low
 *   nibble of the type indicates the width of the length :), the high nibble
 *   the width of the offset.
 * . The first offset is relative to cluster 0, later offsets are relative to
 *   the previous cluster.
 *
 * This function decodes a run. Length is an output parameter, data and cluster
 * are in/out parameters.
 */
int decompress_run(unsigned char **data,int *length,int *cluster,int *ctype)
{
	unsigned char type=*(*data)++;
	*ctype=0;
	switch(type & 0xF)
	{
	case 1: *length=NTFS_GETU8(*data);(*data)++;break;
	case 2: *length=NTFS_GETU16(*data);
		*data+=2;
		break;
	case 3: *length = NTFS_GETU24(*data);
		*data+=3;
		break;
		/* TODO: case 4-8 */
	default:
		ntfs_error("Can't decode run type field %x\n",type);
		return -1;
	}
	switch(type & 0xF0)
	{
	case 0:    *ctype=2;break;
	case 0x10: *cluster+=NTFS_GETS8(*data);(*data)++;break;
	case 0x20: *cluster+=NTFS_GETS16(*data);
		*data+=2;
		break;
	case 0x30: 	*cluster+=NTFS_GETS24(*data);
		*data+=3;
		break;
		/* TODO: case 0x40-0x80 */
	default:
		ntfs_error("Can't decode run type field %x\n",type);
		return -1;
	}
	return 0;
}

/* Reads l bytes of the attribute (attr,name) of ino starting at offset
   on vol into buf. Returns the number of bytes read. 0 is end of attribute,
   -1 is error */
int ntfs_readwrite_attr(ntfs_inode *ino,int type,
	char *name,int offset,ntfs_io *dest,int l)
{
	ntfs_attribute *attr;
	int datasize,resident,rnum,compressed;
	int cluster,s_cluster,vcn,len,chunk,copied;
	int s_vcn,s_vcn_comp;
	int clustersize;

	attr=ntfs_find_attr(ino,type,name);
	if(!attr)
		return -1;
	clustersize=ino->vol->clustersize;
	datasize=DATASIZE(&attr->header);
	if(dest->do_read)
	{
		if(offset>datasize)
			return 0;
		if(offset+l>=datasize)
			l=datasize-offset;
	}else{
		/* if writing beyond end, extend attribute */
	}
	resident=RESIDENT(&attr->header);
	if(resident)
	{
		if(dest->do_read)
			dest->fn_put(dest,attr->d.data+offset,l);
		else
		{
			dest->fn_get(attr->d.data+offset,dest,l);
			update_inode(ino);
		}
		return l;
	}
	compressed=COMPRESSED(&attr->header);
	vcn=0;
	s_vcn = offset/clustersize;
	s_vcn_comp = compressed ? (s_vcn/16)*16 : s_vcn;
	for(rnum=0;vcn+attr->d.r.runlist[rnum].len<=s_vcn_comp;rnum++)
		vcn+=attr->d.r.runlist[rnum].len;
	
	copied=0;
	while(l)
	{
		s_vcn = offset/clustersize;
		s_vcn_comp = compressed ? (s_vcn/16)*16 : s_vcn;
		cluster=attr->d.r.runlist[rnum].cluster;
		len=attr->d.r.runlist[rnum].len;

		if(compressed && cluster==-1)
		{
			/* We are in the middle of a sparse data block */
			char *sparse = ntfs_malloc(512);
			if(!sparse)return -1;
			ntfs_bzero(sparse,512);
			chunk = min((vcn+len)*clustersize-offset,l);
			while(chunk)
			{
				int i=min(chunk,512);
				dest->fn_put(dest,sparse,i);
				chunk-= i;
			}
			/* restore original value */
			chunk = min((vcn+len)*clustersize-offset,l);
			ntfs_free(sparse);
		}else if(compressed)
		{
			char *comp,*comp1;
			char *decomp;
			int got,vcn_comp1;
			int offs1;
			/* FIXME: writing compressed files */
			if(!dest->do_read)
				return -1;
			if(!(comp= ntfs_malloc(16*clustersize)))
				return -1;
			if(!(decomp = ntfs_malloc(16*clustersize)))
			{
				ntfs_free(comp);
				return -1;
			}
			comp1=comp;
			got=0;
			vcn_comp1=s_vcn_comp;
			do{
				ntfs_io io;
				int l1=min(len*clustersize,16*clustersize-got);
				io.do_read=1;
				io.fn_put=ntfs_put;
				io.fn_get=0;
				io.param=comp1;
				if(!ntfs_getput_clusters(ino->vol,cluster+vcn_comp1-vcn,0,
							 l1,&io))
				{
					ntfs_free(comp);ntfs_free(decomp);return -1;
				}
				comp1+=len*clustersize;
				if(l1==len*clustersize){
					rnum++;
					vcn+=len;
					cluster=attr->d.r.runlist[rnum].cluster;
					len=attr->d.r.runlist[rnum].len;
					vcn_comp1=vcn;
				}
				got+=l1;
			}while(cluster!=-1 && got<16*clustersize); /*until 'emtpy' run*/

			offs1 = offset-s_vcn_comp*clustersize;
			if(cluster!=-1){
				/* uncompressible */
				comp1 = comp;
			}else{
				ntfs_decompress(decomp,comp,min(offs1+l,16*clustersize));
				comp1 = decomp;
			}		
			chunk = min(16*clustersize-offs1,l);
			dest->fn_put(dest,comp1+offs1,chunk);
			ntfs_free(comp);
			ntfs_free(decomp);
		}else{
			s_cluster = cluster+s_vcn-vcn;

			chunk=min((vcn+len)*clustersize-offset,l);
			if(!ntfs_getput_clusters(ino->vol,s_cluster,
						 offset-s_vcn*clustersize,chunk,dest))
			{
				ntfs_error("Read error\n");
				return copied;
			}
		}
		l-=chunk;
		copied+=chunk;
		offset+=chunk;
		if(l && offset>=((vcn+len)*clustersize))
		{
			rnum++;
			vcn+=len;
			cluster = attr->d.r.runlist[rnum].cluster;
			len = attr->d.r.runlist[rnum].len;
		}
	}
	return copied;
}

int ntfs_read_attr(ntfs_inode *ino,int attr,
	char *name,int offset,ntfs_io *buf,int l)
{
	buf->do_read=1;
	return ntfs_readwrite_attr(ino,attr,name,offset,buf,l);
}

int ntfs_write_attr(ntfs_inode *ino,int attr,
	char *name,int offset,ntfs_io *buf,int l)
{
	buf->do_read=0;
	return ntfs_readwrite_attr(ino,attr,name,offset,buf,l);
}

int ntfs_vcn_to_lcn(ntfs_inode *ino,int vcn)
{
	int rnum;
	ntfs_attribute *data=ntfs_find_attr(ino,AT_DATA,0);
	/* It's hard to give an error code */
	if(!data)return -1;
	if(RESIDENT(&data->header))return -1;
	if(COMPRESSED(&data->header))return -1;
	if(DATASIZE(&data->header)<vcn*ino->vol->clustersize)return -1;

	for(rnum=0;rnum<data->d.r.len && 
		    vcn>data->d.r.runlist[rnum].len;rnum++)
		vcn-=data->d.r.runlist[rnum].len;
	
	return data->d.r.runlist[rnum].cluster+vcn;
}

/* copy len unicode characters from from to to :) */
void ntfs_uni2ascii(char *to,char *from,int len)
{
	int i;
	for(i=0;i<len;i++)
		to[i]=from[2*i];
	to[i]='\0';
}

/* copy len asci characters from from to to :) */
void ntfs_ascii2uni(short int *to,char *from,int len)
{
	int i;
	for(i=0;i<len;i++)
		to[i]=from[i];
	to[i]=0;
}

/* strncmp for Unicode strings */
int ntfs_uni_strncmp(short int* a,short int *b,int n)
{
	int i;
	for(i=0;i<n;i++)
	{
		if(a[i]<b[i])
			return -1;
		if(b[i]<a[i])
			return 1;
	}
	return 0;
}

/* strncmp between Unicode and ASCII strings */
int ntfs_ua_strncmp(short int* a,char* b,int n)
{
	int i;
	for(i=0;i<n;i++)
	{
		if(a[i]<b[i])
			return -1;
		if(b[i]<a[i])
			return 1;
	}
	return 0;
}

/* Convert the NT UTC (based 1.1.1601, in hundred nanosecond units)
   into Unix UTC (based 1.1.1970, in seconds) */
ntfs_time_t ntfs_ntutc2unixutc(ntfs_time64_t ntutc)
{
/*
 * This is very gross because
 * 1: We must do 64-bit division on a 32-bit machine
 * 2: We can't use libgcc for long long operations in the kernel
 * 3: Floating point math in the kernel would corrupt user data
 */
	const unsigned int D = 10000000;
	unsigned int H = (ntutc >> 32);
	unsigned int L = (unsigned int)ntutc;
	unsigned int numerator2;
	unsigned int lowseconds;
	unsigned int result;

	/* It is best to subtract 0x019db1ded53e8000 first. */
	/* Then the 1601-based date becomes a 1970-based date. */
	if(L < (unsigned)0xd53e8000) H--;
	L -= (unsigned)0xd53e8000;
	H -= (unsigned)0x019db1de;

	/*
	 * Now divide 64-bit numbers on a 32-bit machine :-)
	 * With the subtraction already done, the result fits in 32 bits.
	 * The numerator fits in 56 bits and the denominator fits
	 * in 24 bits, so we can shift by 8 bits to make this work.
	 */

	numerator2  = (H<<8) | (L>>24);
	result      = (numerator2 / D);   /* shifted 24 right!! */
	lowseconds  = result << 24;

	numerator2  = ((numerator2-result*D)<<8) | ((L>>16)&0xff);
	result      = (numerator2 / D);   /* shifted 16 right!! */
	lowseconds |= result << 16;

	numerator2  = ((numerator2-result*D)<<8) | ((L>>8)&0xff);
	result      = (numerator2 / D);   /* shifted 8 right!! */
	lowseconds |= result << 8;

	numerator2  = ((numerator2-result*D)<<8) | (L&0xff);
	result      = (numerator2 / D);   /* not shifted */
	lowseconds |= result;

	return lowseconds;
}

ntfs_time64_t ntfs_unixutc2ntutc (ntfs_time_t t)
{
	return ((t + (ntfs_time64_t)(369*365+89)*24*3600) * 10000000);
}

static int allocate_store(ntfs_volume *vol,ntfs_disk_inode *store,int count)
{
	int i;
	if(store->count>count)
		return 0;
	if(store->size<count){
		ntfs_mft_record* n=ntfs_malloc((count+4)*sizeof(ntfs_mft_record));
		if(!n)
			return ENOMEM;
		if(store->size){
			for(i=0;i<store->size;i++)
				n[i]=store->records[i];
			ntfs_free(store->records);
		}
		store->size=count+4;
		store->records=n;
	}
	for(i=store->count;i<count;i++){
		store->records[i].record=ntfs_malloc(vol->mft_recordsize);
		if(!store->records[i].record)
			return ENOMEM;
		store->count++;
	}
	return 0;
}

static void deallocate_store(ntfs_disk_inode* store)
{
	int i;
	for(i=0;i<store->count;i++)
		ntfs_free(store->records[i].record);
	ntfs_free(store->records);
	store->count=store->size=0;
	store->records=0;
}

int layout_runs(ntfs_attribute *attr,char* rec,int* offs,int size)
{
	int i,cluster,rclus,len,offset,coffs;
	ntfs_runlist *rl=attr->d.r.runlist;
	cluster=0;
	offset=*offs;
	for(i=0;i<attr->d.r.len;i++){
		rclus=rl[i].cluster-cluster;
		len=rl[i].len;
		rec[offset]=0;
		if(offset+8>size)
			return E2BIG; /* it might still fit, but this simplifies testing */
		if(len<0x100){
			*(rec+offset)|=1;
			NTFS_PUTU8(rec+offset+1,len);
			coffs=2;
		}else if(len<0x10000){
			*(rec+offset)|=2;
			NTFS_PUTU16(rec+offset+1,len);
			coffs=3;
		}else if(len<0x1000000){
			*(rec+offset)|=3;
			NTFS_PUTU24(rec+offset+1,len);
			coffs=4;
		}else{
			*(rec+offset)|=4;
			NTFS_PUTU32(rec+offset+1,len);
			coffs=5;
		}
    
		if(rl[i].cluster==0) /*compressed run*/
			/*nothing*/;
		else if(rclus>-0x80 && rclus<0x7F){
			*(rec+offset)|=0x10;
			NTFS_PUTS8(rec+offset+coffs,rclus);
			coffs+=1;
		}else if(rclus>-0x8000 && rclus<0x7FFF){
			*(rec+offset)|=0x20;
			NTFS_PUTS16(rec+offset+coffs,rclus);
			coffs+=2;
		}else if(rclus>-0x800000 && rclus<0x7FFFFF){
			*(rec+offset)|=0x30;
			NTFS_PUTS24(rec+offset+coffs,rclus);
			coffs+=3;
		}else{
			*(rec+offset)|=0x40;
			NTFS_PUTS32(rec+offset+coffs,rclus);
			coffs+=4;
		}
		offset+=coffs;
		if(rl[i].cluster)
			cluster=rl[i].cluster;
	}
	*offs=offset;
	return 0;
}

static void count_runs(ntfs_attribute *attr,char *buf)
{
	int first,count,last,i;
	first=0;
	for(i=0,count=0;i<attr->d.r.len;i++)
		count+=attr->d.r.runlist[i].len;
	last=first+count-1;

	NTFS_PUTU32(buf+0x10,first);
	NTFS_PUTU32(buf+0x18,last);
} 

/* Try to layout ino into store. Return 0 on success,
   E2BIG if it does not fit, 
   ENOMEM if memory allocation problem,
   EOPNOTSUP if beyound our capabilities 
*/
int layout_inode(ntfs_inode *ino,ntfs_disk_inode *store)
{
	int offset,next,i;
	ntfs_attribute *attr;
	unsigned char *rec;
	int size,asize,aoffset;
	int error;

	if(ino->record_count>1)
	{
		ntfs_error("layout_inode: attribute lists not supported\n");
		return EOPNOTSUPP;
	}
	error=allocate_store(ino->vol,store,1);
	if(error)
		return error;
	rec=store->records[0].record;
	size=ino->vol->mft_recordsize;
	store->records[0].recno=ino->records[0];
	/* copy header */
	offset=NTFS_GETU16(ino->attr+0x14);
	ntfs_memcpy(rec,ino->attr,offset);
	for(i=0;i<ino->attr_count;i++){
		attr=ino->attrs+i;
		/* copy attribute header */
		ntfs_memcpy(rec+offset,attr->header,
			    min(sizeof(attr->header),size-offset)); /* consider overrun */
		if(attr->namelen)
			/* named attributes are added later */
			return EOPNOTSUPP;
		/* FIXME: assign attribute ID??? */
		if(RESIDENT(&attr->header)){
			asize=DATASIZE(&attr->header);
			aoffset=NTFS_GETU16(rec+offset+0x14);
			if(offset+aoffset+asize>size)
				return E2BIG;
			ntfs_memcpy(rec+offset+aoffset,attr->d.data,asize);
			next=offset+aoffset+asize;
		}else{
			count_runs(attr,rec+offset);
			aoffset=NTFS_GETU16(rec+offset+0x20);
			next=offset+aoffset;
			error=layout_runs(attr,rec,&next,size);
			if(error)
				return error;
		}
		next=(next+7) & ~7; /* align to DWORD */
		NTFS_PUTU16(rec+offset+4,next-offset);
		offset=next;
	}
	/* terminating attribute */
	if(offset+8<size){
		NTFS_PUTU32(rec+offset,0xFFFFFFFF);
		offset+=4;
		NTFS_PUTU32(rec+offset,0);
		offset+=4;
	}else
		return E2BIG;
	NTFS_PUTU32(rec+0x18,offset);
	return 0;
}
  

int update_inode(ntfs_inode *ino)
{
	int error;
	ntfs_disk_inode store;
	ntfs_io io;
	int i,l;

	store.count=store.size=0;
	store.records=0;
	error=layout_inode(ino,&store);
	if(error==E2BIG){
		/* should try:
		   make attributes non-resident
		   introduce extension records
		   */
		ntfs_error("cannot handle saving inode %x\n",ino->i_number);
		deallocate_store(&store);
		return EOPNOTSUPP;
	}
	if(error){
		deallocate_store(&store);
		return error;
	}
	io.fn_get=ntfs_get;
	io.fn_put=0;
	for(i=0;i<store.count;i++){
		ntfs_insert_fixups(store.records[i].record,ino->vol->blocksize);
		io.param=store.records[i].record;
		/* FIXME: is this the right way? */
		l=ntfs_write_attr(ino->vol->mft_ino,AT_DATA,0,
				  store.records[i].recno*ino->vol->mft_recordsize,
				  &io,ino->vol->mft_recordsize);
		if(l!=ino->vol->mft_recordsize){
			/* big trouble, partially written file */
			ntfs_error("Please unmount: write error in inode %x\n",ino->i_number);
			deallocate_store(&store);
			/* FIXME: error code */
			return EIO;
		}
	}
	return 0;
}

void ntfs_decompress(unsigned char* dest,unsigned char*src,ntfs_size_t l)
{
	int head;
	int copied=0;
	unsigned char *stop;
	int bits;
	int tag=0;
	int clear_pos;
	while(1)
	{
		head = NTFS_GETU16(src) & 0xFFF;
		src += 2;
		stop = src+head;
		bits = 0;
		clear_pos=0;
		if(head==0xFFF) /* uncompressible */
		{
			ntfs_memcpy(dest,src,0x1000);
			dest+=0x1000;
			copied+=0x1000;
			if(l==copied)
				return;
			continue;
		}
		while(src<=stop)
		{
			if(clear_pos>4096)
			{
				ntfs_error("Error 1 in decompress\n");
				return;
			}
			if(!bits){
				tag=NTFS_GETU8(src);
				bits=8;
				src++;
				if(src>stop)
					break;
			}
			if(tag & 1){
				int i,len,delta,code,lmask,dshift;
				code = NTFS_GETU16(src);
				src+=2;
				if(!clear_pos)
				{
					ntfs_error("Error 2 in decompress\n");
					return;
				}
				for(i=clear_pos-1,lmask=0xFFF,dshift=12;i>=0x10;i>>=1)
				{
					lmask >>= 1;
					dshift--;
				}
				delta = code >> dshift;
				len = (code & lmask) + 3;
				for(i=0; i<len; i++)
				{
					dest[clear_pos]=dest[clear_pos-delta-1];
					clear_pos++;
					copied++;
					if(copied==l)
						return;
				}
			}else{
				dest[clear_pos++]=NTFS_GETU8(src);
				src++;
				copied++;
				if(copied==l)
					return;
			}
			tag>>=1;
			bits--;
		}
		dest+=clear_pos;
	}
}

/* Caveat: No range checking in either ntfs_set_bit or ntfs_clear_bit */
void ntfs_set_bit (unsigned char *byte, int bit)
{
	byte += (bit >> 3);
	bit &= 7;
	*byte |= (1 << bit);
}

void ntfs_clear_bit (unsigned char *byte, int bit)
{
	byte += (bit >> 3);
	bit &= 7;
	*byte &= ~(1 << bit);
}

/* We have to skip the 16 metafiles and the 8 reserved entries */
int ntfs_new_inode (unsigned char *buffer, int length, int* result)
{
	int byte;
	int bit;
	unsigned char value;

	for (byte = 3; byte < length; byte++)
	{
		value = buffer[byte];
		for (bit = 0; bit < 8; bit++, value >>= 1)
		{
			if (!(value & 1)){
				*result=byte*8+bit;
				return 0;
			}
		}
	}
	/* There is no free space.  We must first extend the MFT. */
	return EOPNOTSUPP;
}

static int add_mft_header (ntfs_inode *ino)
{
	unsigned short fixup = 1;
	int fixup_count;
	int fixup_offset;
	int attr_offset;
	int record_size;
	unsigned char* mft;
	ntfs_volume *vol=ino->vol;

	fixup_offset = 0x2a;
	
	record_size = vol->mft_recordsize;
	fixup_count = record_size / vol->blocksize + 1;
	attr_offset = (0x2a + (2 * fixup_count) + 7) & ~7;

	mft=ino->attr;

	ntfs_bzero (mft, record_size);

	NTFS_PUTU32(mft + 0x00, 0x454c4946);	     /* file */
	NTFS_PUTU16(mft + 0x04, 0x2a);		     /* offset to fixup */
	NTFS_PUTU16(mft + 0x06, fixup_count);	     /* Number of fixups */
	/* FIXME: sequence number. Regis recommended 5. */
	NTFS_PUTU16(mft + 0x10, 5);
	NTFS_PUTU16(mft + 0x12, 1);                  /* hard link count */
	NTFS_PUTU16(mft + 0x14, attr_offset);	     /* Offset to attributes */
	NTFS_PUTU16(mft + 0x16, 1);                  /*FIXME: flags ?? */
	NTFS_PUTU32(mft + 0x18, attr_offset + 0x08);	/* In use */
	NTFS_PUTU32(mft + 0x1c, record_size);	     /* Total size */

	NTFS_PUTU16(mft + fixup_offset, fixup);		/* Fixup word */
	NTFS_PUTU32(mft + attr_offset, 0xffffffff);	/* End marker */
	return 0;
}

/* We need 0x48 bytes in total */
static int add_standard_information (ntfs_inode *ino)
{
	ntfs_time64_t now;
	char data[0x30];
	char *position=data;
	int error;

	now = ntfs_now();
	NTFS_PUTU64(position + 0x00, now);		/* File creation */
	NTFS_PUTU64(position + 0x08, now);		/* Last modification */
	NTFS_PUTU64(position + 0x10, now);		/* Last mod for MFT */
	NTFS_PUTU64(position + 0x18, now);		/* Last access */

	NTFS_PUTU64(position + 0x20, 0x00);		/* MSDOS file perms */
	NTFS_PUTU64(position + 0x28, 0);               /* unknown */
	error=ntfs_create_attr(ino,AT_STANDARD_INFORMATION,0,data,sizeof(data));

	return error;
}

static int add_filename (ntfs_inode* ino, ntfs_inode* dir, 
			 const unsigned char *filename, int length)
{
	unsigned char   *position;
	unsigned int    size;
	ntfs_time64_t   now;
	int		count;
	int error;
	unsigned char* data;
	ntfs_u64 ino_dir;

	/* work out the size */
	size = 0x42 + 2 * length;
	data = ntfs_malloc(size);
	ntfs_bzero(data,size);

	/* search for a position */
	position = data;

	ino_dir=dir->i_number;
	ino_dir|= ((ntfs_u64)(*(unsigned short*)(dir->attr+0x10)))<<48;
	NTFS_PUTU64(position, ino_dir);	/* Inode num of dir */

	now = ntfs_now();
	NTFS_PUTU64(position + 0x08, now);		/* File creation */
	NTFS_PUTU64(position + 0x10, now);		/* Last modification */
	NTFS_PUTU64(position + 0x18, now);		/* Last mod for MFT */
	NTFS_PUTU64(position + 0x20, now);		/* Last access */

	/* Don't know */
	NTFS_PUTU8(position+0x38, 0x20);  /*archive*/

	NTFS_PUTU8(position + 0x40, length);	      /* Filename length */
	NTFS_PUTU8(position + 0x41, 0x01);	      /* File type (Win32) */

	position += 0x42;
	for (count = 0; count < length; count++)
	{
		NTFS_PUTU16(position + 2 * count, filename[count]);
	}

	error=ntfs_create_attr(ino,AT_FILE_NAME,0,data,size);
	ntfs_free(data);

	return error;
}

int add_security (ntfs_inode* ino, ntfs_inode* dir)
{
	int error;
	char *buf;
	int size;
	ntfs_attribute* attr;
	ntfs_io io;

	attr=ntfs_find_attr(dir,AT_SECURITY_DESCRIPTOR,0);
	if(!attr)
		return EOPNOTSUPP; /* need security in directory */
	size = DATASIZE(&(attr->header));
	if(size>512)
		return EOPNOTSUPP;
	buf=ntfs_malloc(size);
	if(!buf)
		return ENOMEM;
	io.fn_get=ntfs_get;
	io.fn_put=ntfs_put;
	io.param=buf;
	if(ntfs_read_attr(dir,AT_SECURITY_DESCRIPTOR,0,0,&io,size)!=size){
		ntfs_free(buf);
		return EIO; /* FIXME */
	}
	/* FIXME: consider ACL inheritance */
	error=ntfs_create_attr(ino,AT_SECURITY_DESCRIPTOR,0,buf,size);
	ntfs_free(buf);
	return error;
}

static int add_data (ntfs_inode* ino, unsigned char *data, int length)
{
	int error;
	error=ntfs_create_attr(ino,AT_DATA,0,data,length);
	return error;
}


/* We _could_ use 'dir' to help optimise inode allocation */
int ntfs_alloc_inode (ntfs_inode*dir, ntfs_inode *result, 
		      char *filename, int namelen)
{
	unsigned char buffer[512];
	ntfs_io io;
	int size;
	int error;
	ntfs_volume* vol=dir->vol;

	io.fn_put = ntfs_put;
	io.fn_get = ntfs_get;
	io.param  = buffer;

	size = ntfs_read_attr (vol->mft_ino, AT_BITMAP, 0, 0, &io, 512);
	if (size < 1)
		return (0);

	error=ntfs_new_inode (buffer, size, &(result->i_number));
	if(error)
	{
		ntfs_error ("ntfs_get_empty_inode: no free inodes\n");
		return error;
	}

	io.param = buffer;
	ntfs_set_bit (buffer, result->i_number);
 	size = ntfs_write_attr (vol->mft_ino, AT_BITMAP, 0, 0, &io, size);
	if (size < 1)
		return ENOSPC;
	result->vol=vol;
	result->attr=ntfs_malloc(vol->mft_recordsize);
	result->attr_count=0;
	result->attrs=0;
	result->record_count=1;
	result->records=ntfs_malloc(8*sizeof(int));
	result->records[0]=result->i_number;
	error=add_mft_header(result);
	if(error)
		return error;
	error=add_standard_information(result);
	if(error)
		return error;
	error=add_filename(result,dir,filename,namelen);
	if(error)
		return error;
	error=add_security(result,dir);
	/*FIXME: check error */
	error=add_data(result,0,0);
	if(error)
		return error;
	return 0;
}

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
